/*
 * (c) 2008-2009 Adam Lackorzynski <adam@os.inf.tu-dresden.de>
 *     economic rights: Technische Universität Dresden (Germany)
 * Copyright (C) 2017, 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This file is part of TUD:OS and distributed under the terms of the
 * GNU General Public License 2.
 * Please see the COPYING-GPL-2 file for details.
 */
/*
 * Access the HDMI I2C peripheral on the MIPS Creator CI20 board.
 */

#include <l4/devices/cpm-jz4780.h>
#include <l4/devices/gpio-jz4780.h>
#include <l4/devices/hdmi-jz4780.h>
#include <l4/devices/memory.h>
#include <l4/io/io.h>
#include <l4/libedid/edid.h>
#include <l4/re/env.h>
#include <l4/re/c/util/cap_alloc.h>
#include <l4/sys/factory.h>
#include <l4/sys/icu.h>
#include <l4/sys/ipc.h>
#include <l4/sys/irq.h>
#include <l4/sys/rcv_endpoint.h>
#include <l4/vbus/vbus.h>
#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>



enum {
  DDCSCL = 24,  /* via PORTF */
  DDCSDA = 25,  /* via PORTF */
};



/* Device and resource discovery. */

static long item_in_range(long start, long end, long index)
{
  if (start < end)
    return start + index;
  else
    return start - index;
}

static void show_timings(uint8_t *buf)
{
  unsigned int width, height;

  /* Attempt to decode EDID information. */

  libedid_prefered_resolution(buf, &width, &height);
  printf("Preferred resolution: %d x %d\n", width, height);

  libedid_dump_standard_timings(buf);
}

int main(void)
{
  long err;

  /* Buffer for EDID data. */

  uint8_t edid[128];
  unsigned int length = 0, i;

  /* Version details. */

  uint8_t hdmi_major;
  uint16_t hdmi_minor;

  /* Peripheral memory. */

  l4_addr_t cpm_base = 0, cpm_base_end = 0;
  l4_addr_t gpio_base = 0, gpio_base_end = 0;
  l4_addr_t hdmi_base = 0, hdmi_base_end = 0;
  l4_addr_t port_f, port_f_end;

  /* Peripheral abstractions. */

  void *cpm;
  void *gpio_port_f;
  void *hdmi;

  /* Access to IRQs. */

  l4_uint32_t hdmi_irq_start = 0, hdmi_irq_end = 0;
  l4_cap_idx_t icucap, irqcap;

  irqcap = l4re_util_cap_alloc();
  icucap = l4re_env_get_cap("icu");

  if (l4_is_invalid_cap(icucap))
  {
    printf("No 'icu' capability available in the virtual bus.\n");
    return 1;
  }

  if (l4_is_invalid_cap(irqcap))
  {
    printf("Capabilities not available for interrupts.\n");
    return 1;
  }

  /* Obtain resource details describing the interrupt for HDMI I2C. */

  printf("Access IRQ...\n");

  if (get_irq("jz4780-hdmi", &hdmi_irq_start, &hdmi_irq_end) < 0)
    return 1;

  printf("IRQ range at %d...%d.\n", hdmi_irq_start, hdmi_irq_end);

  /* Create interrupt objects. */

  err = l4_error(l4_factory_create_irq(l4re_global_env->factory, irqcap));

  if (err)
  {
    printf("Could not create IRQ object: %lx\n", err);
    return 1;
  }

  /* Bind interrupt objects to IRQ numbers. Here, the first HDMI interrupt is
     bound, this being the general HDMI interrupt. */

  err = l4_error(l4_icu_bind(icucap,
                             item_in_range(hdmi_irq_start, hdmi_irq_end, 0),
                             irqcap));

  if (err)
  {
    printf("Could not bind IRQ to the ICU: %ld\n", err);
    return 1;
  }

  /* Attach ourselves to the interrupt handler. */

  err = l4_error(l4_rcv_ep_bind_thread(irqcap, l4re_env()->main_thread, 0));

  if (err)
  {
    printf("Could not attach to IRQs: %ld\n", err);
    return 1;
  }

  /* Obtain resource details describing I/O memory. */

  printf("Access CPM...\n");

  if (get_memory("jz4780-cpm", &cpm_base, &cpm_base_end) < 0)
    return 1;

  printf("CPM at 0x%lx...0x%lx.\n", cpm_base, cpm_base_end);

  printf("Access GPIO...\n");

  if (get_memory("jz4780-gpio", &gpio_base, &gpio_base_end) < 0)
    return 1;

  printf("GPIO at 0x%lx...0x%lx.\n", gpio_base, gpio_base_end);

  printf("Access HDMI...\n");

  if (get_memory("jz4780-hdmi", &hdmi_base, &hdmi_base_end) < 0)
    return 1;

  printf("HDMI at 0x%lx...0x%lx.\n", hdmi_base, hdmi_base_end);

  /* Obtain CPM object. */

  cpm = jz4780_cpm_init(cpm_base);

  printf("VPLL frequency: %d\n", jz4780_cpm_get_vpll_frequency(cpm));
  printf("HDMI divider: %d\n", jz4780_cpm_get_hdmi_divider(cpm));
  printf("HDMI frequency: %d\n", jz4780_cpm_get_hdmi_frequency(cpm));

  jz4780_cpm_stop_hdmi(cpm);
  jz4780_cpm_set_hdmi_frequency(cpm, 27000000);

  printf("HDMI divider: %d\n", jz4780_cpm_get_hdmi_divider(cpm));
  printf("HDMI frequency: %d\n", jz4780_cpm_get_hdmi_frequency(cpm));

  jz4780_cpm_start_hdmi(cpm);

  /* Configure pins. */

  port_f = gpio_base + 0x500;
  port_f_end = port_f + 0x100;

  printf("PORTF at 0x%lx...0x%lx.\n", port_f, port_f_end);

  gpio_port_f = jz4780_gpio_init(port_f, port_f_end, 32, 0x7fa7f00f, 0x00580ff0);

  printf("Set up GPIO pins...\n");

  jz4780_gpio_config_pad(gpio_port_f, DDCSCL, Function_alt, 0);
  jz4780_gpio_config_pad(gpio_port_f, DDCSDA, Function_alt, 0);

  /* Obtain HDMI reference. */

  printf("Set up HDMI...\n");

  hdmi = jz4780_hdmi_init(hdmi_base, hdmi_base_end, irqcap);

  printf("Read version...\n");

  jz4780_hdmi_get_version(hdmi, &hdmi_major, &hdmi_minor);

  printf("HDMI version is %x.%03x\n", hdmi_major, hdmi_minor);

  printf("Read EDID...\n");

  jz4780_hdmi_i2c_set_address(hdmi, 0x50);
  length = jz4780_hdmi_i2c_read(hdmi, edid, 128);

  if (length)
  {
    for (i = 0; i < length; i++)
      printf("%02x%c", edid[i], ((i % 16) != 15) ? ' ' : '\n');
  }

  show_timings(edid);

  /* Detach from the interrupt. */

  err = l4_error(l4_irq_detach(irqcap));

  if (err)
    printf("Error detaching from IRQ: %ld\n", err);

  printf("Done.\n");

  return 0;
}
